#!/usr/bin/python3
# subscribe to the machine model, and periodically run functions from daemon.py

###################################################
# configuration definitions
# these values can be changed to affect behaviour

# this is initial frequency of running the loop() function
# which is also the frequency of updating the object model
# any of the daemon functions can return a new value for interval
loop_interval = 30

# daemon file location directory and file name
# printer sys folder location is /opt/dsf/sd/sys
# note, file name MUST end in '.py'
daemon_d = "/opt/dsf/sd/sys"
daemon_f = "daemon.py"

# to see messages when running as a plugin, at pi command prompt:
# journalctl -u duetpluginservice -f

# print some running commentary
# if verbose is False, the only thing reported is reading a new daemon.py file
verbose = False

# if vTS is true, commentary will have timestamps (to 1mS)
vTS = False

# end of configuration definitions
###################################################

# imports
# note the dameon.py file is imported programmatically further down
# since it is monitored for changes (and coming and going)

import json
import time
import sys
import pathlib
import importlib
import re
import inspect
from dsf.connections import SubscribeConnection, SubscriptionMode

# global variables

# object model
om = {}

# whether we have ever succesfully imported the daemon.py file
dpy_imported = False
# modified time of the daemon.py file - initially at the epoch!
dpy_modtime=0



###################################################
# function definitions

# update dictionary-of-dicts-and-lists a with a merge patch b
# values in b over-write values with same key tree in a
# notes
#   + this function recurses
#   + dsf-python patches are just a json string so need to be turned into a dict
#     BEFORE feeding to this function with the object model dict
#   + if a and b are both lists and b is shorter than a, a will be truncated
#     but this behaviour can be changed below - search 'truncated'
# debug=True draws a diagram
# debug=True and verbose=True reports input to the function on every entry
def patch(a, b, path=None, debug=False, verbose=False):
    if path is None:
        path = []
    if debug and (len(path) == 0 or verbose):
        print('patch: ' + str(b))
        if verbose:
            print(' into: ' + str(a))
            print(' at :  ' + str(path))

    # if both a and b are dicts work through the keys
    if isinstance(a, dict) and isinstance(b, dict):
        for key in b:
            if debug:
                # debug draws a sort of diagram indicating nesting depth
                print('   ', end='')
                for _ in path:
                    print(' > ', end='')
                print(key)
            if key in a:
                # the element in the patch already exists in the original
                if ((isinstance(a[key], dict) and isinstance(b[key], dict))
                        or (isinstance(a[key], list) and isinstance(b[key], list))):
                    # a[key] and b[key] are both dicts or both lists, recurse
                    a[key] = patch(a[key], b[key], path + [str(key)], debug, verbose)
                else:
                    # mixed types, so treat as leaf
                    if debug:
                        for _ in path:
                            print('   ', end='')
                        print('    = ' + str(b[key]))
                    a[key] = b[key]
            else:
                # the element in the patch is not in the original
                # treat as a leaf, but note either could actually be a dict or list
                # e.g. a[key] could have been a single value and we now overwrite with a new dict
                # or it could have been a dict and now we've overwritten a single scalar
                if debug:
                    for _ in path:
                        print('   ', end='')
                    print('    = ' + str(b[key]))
                a[key] = b[key]

    # if both a and b are lists we enumerate and work through the values
    elif isinstance(a, list) and isinstance(b, list):
        # if b is shorter than a, should a be truncated or should the trailing values be unaltered?
        # The following line assumes that trailing values not found in b are truncated from a
        # remove this line if desired behaviour is that trailing values in a are left in place
        # note that would mean that a list will grow but never shrink
        a = a[0:len(b)]
        # if b is longer than a I could use a try: except IndexError: below
        # but that requires putting an entire recursion of this function inside the try
        # which makes it difficult to track where bad things are happening, so instead
        # a is padded to be as long as b in advance
        a.extend([None] * (len(b) - len(a)))
        for idx, value in enumerate(b):
            # need a new diagram nesting line
            if debug:
                print('   ', end='')
                for _ in path:
                    print(' > ', end='')
                # put [] around index to identify we're in a list
                print('[' + str(idx) + ']')
            a[idx] = patch(a[idx], b[idx], path + [str(idx)], debug, verbose)

    else:
        # a and b are different types
        if debug:
            for _ in path:
                print('   ', end='')
            print(' = ' + str(b))
        a = b

    # all done
    sys.stdout.flush()
    return a


#################################################
# main
#################################################

if __name__ == '__main__':

    # establish the dsf-python connection, with some re-tries just in case
    for attempt in range(5):
        try:
            if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'try subscribe connection', flush=True)
            subscribe_connection = SubscribeConnection(SubscriptionMode.PATCH, debug=False)
            subscribe_connection.connect()
        except:
            time.sleep(3)
        else:
            if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'got subscribe connection', flush=True)
            break
    else:
        # ran to end of our attempts and didn't break out so still not established
        # make one last attempt after a longer pause
        time.sleep(15)
        subscribe_connection = SubscribeConnection(SubscriptionMode.PATCH, debug=False)
        subscribe_connection.connect()

    try:
        # get the complete object model once
        # note we cast it to a string in order to feed it in to json.loads
        om = json.loads(str(subscribe_connection.get_object_model()))
        if verbose: 
            print(("{:.6f} ".format(time.time()) if vTS else '') + 'got object model', flush=True)
            # report what boards we have found
            print('  boards:')
            for b in om['boards']:
                print('   ',b['uniqueId'],':',b['name'], flush=True)

        # we want a regex to match the creation of the various functions in teh daemon.py file
        fn_setup_pattern = re.compile('^\s*def\s+(setup)\s*\(.+\):')
        fn_loop_pattern = re.compile('^\s*def\s+(loop)\s*\(.+\):')
        fn_loopn_pattern = re.compile('^\s*def\s+(loop(\d+)\S*)\s*\(.+\):')

        # main loop which runs on each multiple of loop_interval seconds since the epoch
        loop_count = 0
        while True:
            time.sleep(loop_interval - time.time() % loop_interval)
            if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'waking', flush=True)

            # generally get and apply object model patch twice - seems to kick a more up-to-date update?
            # if refreshing at 2 seconds or less however, just the once
            for n in range(2 if loop_interval>2 else 1):
                # get object model update as a string
                om_u_str = subscribe_connection.get_object_model_patch()
                # convert to a  dict
                om_update = json.loads(om_u_str)
                # apply to the saved object model
                patch(om, om_update)
            if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'object model updated', flush=True)

            # look for the daemon.py file and get its modified time
            try:
                new_modtime = pathlib.Path(daemon_d+'/'+daemon_f).stat().st_mtime
            except FileNotFoundError:
                # looks like daemon.py has gone away (if it was ever here)
                if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + daemon_f + ' file not found', flush=True)
                # we will skip trying to run it and wait until next time
                modtime = 0
                dpy_modtime = 0
                # if we currently have daemon file imported, drop all the functions from it
                # we do this becasue if old daemon.py defined a function that new daemon.py 
                # does not have, that function will still be cached
                # that could be relevant if we start searching through the list of functions
                # see 'caveats' under https://docs.python.org/3/library/importlib.html#importlib.reload
                if dpy_imported:
                    for (f,ff) in inspect.getmembers(daemon, inspect.isfunction):
                        delattr(daemon,f)
                    dpy_imported = False
            else:
                # check if the daemon.py file has changed (or appeared)
                if (new_modtime > dpy_modtime):
                    # not very pythonic, but we check if the file has already been imported
                    # to determine if we need an import_module or a reload
                    if dpy_imported:
                        print(("{:.6f} ".format(time.time()) if vTS else '') + daemon_f + ' file has changed, reload', flush=True)
                        # ditch teh current functions (see note why above)
                        for (f,ff) in inspect.getmembers(daemon, inspect.isfunction):
                            delattr(daemon,f)
                        # reload previously specified module
                        importlib.reload(daemon)
                    else:
                        print(("{:.6f} ".format(time.time()) if vTS else '') + daemon_f + ' file has appeared, open', flush=True)
                        sys.path.append(daemon_d)
                        daemon = importlib.import_module(daemon_f[:-3])
                        dpy_imported = True
                    # record modification time of this version of the file
                    dpy_modtime=pathlib.Path(daemon_d+'/'+daemon_f).stat().st_mtime
                    # now think about what loop functions we have
                    # assemble a dictionary where key is name of the function and value the interval count
                    functions = {}
                    # record teh existence of plain setup() and loop() functions
                    has_setup_fn = False
                    has_loop_fn = False
                    # look for other functions with integer suffixes
                    # note we read the source, we cannot just look for functions because we want to 
                    # preserve the order in which the functions appear in the source file
                    f = open(daemon_d+'/'+daemon_f, 'r')
                    for l in f:
                        # patterns have previously been compiled as a re
                        if fn_setup_pattern.match(l):
                            print(("{:.6f} ".format(time.time()) if vTS else '') + 'found setup() function', flush=True)
                            has_setup_fn = True
                        if fn_loop_pattern.match(l):
                            print(("{:.6f} ".format(time.time()) if vTS else '') + 'found loop() function', flush=True)
                            has_loop_fn = True
                            functions = {'loop':1, **functions}
                        m = fn_loopn_pattern.match(l)
                        if m:
                            print(("{:.6f} ".format(time.time()) if vTS else '') + 'found ' + m.group(1) +'() function', flush=True)
                            functions[m.group(1)] = int(m.group(2))
                    f.close()
                    # report to user if required
                    if verbose:
                        print(("{:.6f} ".format(time.time()) if vTS else '') + 'functions to run', flush=True)
                        if has_setup_fn: print('    setup [once]')
                        for f in functions.keys():
                            print('    ' + f + ' [every ' + str(functions[f]) + ' intervals]', flush=True)

                    # run setup function
                    if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'calling setup()', flush=True)
                    rtn = daemon.setup(om)
                    # if the return value makes sense as a number, we set the loop interval to that
                    try:
                        loop_interval = float(rtn)
                    except (TypeError,ValueError):
                        pass
                    if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'setup() completed', flush=True)
                    # initialise the loop counter
                    loop_count = 0

                # run the loop functions
                if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + str(loop_count) , flush=True)
                for f in functions.keys():
                    if (loop_count % functions[f]) == 0:
                        if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'calling ' + f + '()', flush=True)
                        rtn = getattr(daemon,f)(om)
                        # if the return value makes sense as a number, we set loop interval to that
                        try:
                            loop_interval = float(rtn)
                        except (TypeError,ValueError):
                            pass

                if verbose: print(("{:.6f} ".format(time.time()) if vTS else '') + 'loop functions completed - sleeping', flush=True)
                loop_count += 1

    finally:
        subscribe_connection.close()
